# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin create_all_node_db}

Create an All Node Database
###########################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

all_node_database
*****************
is a python string containing the name of the
:ref:`all_node_db-name` that is created by this call.
This argument can't be ``None``.

option_all
**********
This argument can't be ``None``.
It is a ``dict`` with a key equal to each
:ref:`option_all_table@Table Format@option_name` that appears in the
option_all table.
The value corresponding to the key is the
:ref:`option_all_table@Table Format@option_value`
in the same row of the option_all_table.
Note that keys must have type ``str`` and all the values will be converted
to ``str``.

fit_goal_table
**************
This specifies the nodes in the :ref:`fit_goal_table-name`.
It must be a ``list`` of ``dict`` representation of the
:ref:`fit_goal_table-name` with one key for each ``dict``.
If the key is node_id (node_name) the corresponding value is
the ``int`` ( ``str`` ) representation of the node.
If this argument is ``None``, the fit_goal table is empty.

split_reference_table
*********************
This specifies the possible reference values for the splitting covariate.
It must be a ``list`` of ``dict`` representation of the
:ref:`split_reference_table-name`.
If this argument is ``None``, the split_reference table is empty.

node_split_table
****************
This specifies the nodes at which the cascade will split by the value
of the splitting covariate.
It must be a ``list`` of ``dict`` representation of the
:ref:`node_split_table-name` with one key for each ``dict``.
If the key is node_id (node_name) the corresponding value is
the ``int`` ( ``str`` ) representation of the node.
If this argument is ``None``, the node_split table is empty.

mulcov_freeze_table
*******************
This specifies the jobs at which the cascade will freeze specific
covariate multipliers.
If this argument is ``None``, the node_split table is empty.
Otherwise, it must be a ``list`` of ``dict`` representation of the
:ref:`mulcov_freeze_table-name` .
We use *row* below to refer to a ``dict`` that is an entry in this ``list`` .

fit_node_id, fit_node_name
==========================
#. If *row*\ ``["fit_node_id"]`` must be ``None``
   if and only if the :ref:`split_reference_table-name` is empty.
#. Otherwise if *row*\ ``["fit_node_id"]`` exists it must be an
   ``int`` representation of the fit_node_id for this job.
#. Otherwise *row*\ ``["fit_node_name"]`` is a
   ``str`` representation of the corresponding node name.

split_reference_id
==================
The value *row*\ ``["split_reference_id"]`` is an
``int`` representation of the split_reference_id for this job or
*row*\ ``["split_reference_name"]`` is a
``str`` representation of the corresponding split reference name.

mulcov_id
=========
The value *row*\ ``["mulcov_id"]`` is an
``int`` representation of the mulcov_id for the covariate multiplier
that is frozen.

omega_grid
**********
is a dictionary with two keys ``age`` and ``time``.
If this argument is ``None``,
the :ref:`omega_grid@omega_age_grid Table` and
the :ref:`omega_grid@omega_time_grid Table` are empty.

age
===
The value *omega_grid*\ [``age``] is a list containing the
age_id values for the omega_grid.
These are indices in the
:ref:`glossary@root_database` age table.
We use the notation *n_omega_age* for the length of the age list.

time
====
The value *omega_grid*\ [``time``] is a list containing the
time_id values for the omega_grid.
These are indices in the *root_database* time table.
We use the notation *n_omega_time* for the length of the time list.

omega_data
**********
This is a python dictionary with a key for each node name
for the :ref:`glossary@root_node` and its descendant.
The value *omega_data[node_name]* is a list.
For each *k*, *omega_data[node_name][k]* is a list.
For *i* equal 0, ..., *n_omega_age*-1
and *j* equal 0, ..., *n_omega_time*-1,

| |tab| *omega_data[node_name][k][ i * n_omega_time + j ]*

is the value of *omega* at the specified node,
the age corresponding to index *i* in *omega_grid*\ [``age``],
and time corresponding to index *j* in *omega_grid*\ [``time``].
If split_reference table is empty, *k* is zero.
Otherwise, let *n_split* be the length of the split_reference table.
For *k* equal 0, ... , *n_split*-1, it specifies the value of
:ref:`split_reference_table@split_reference_id` for the covariate value.

default
=======
The *omega_data* argument is ``None`` if and only if *omega_grid* is ``None``.
If *omega_data* is ``None`` the
:ref:`omega_all@omega_all Table` and
:ref:`omega_all@omega_index Table` will be empty.

cov_reference_table
*******************
This must be ``None`` or a ``list`` of ``dict`` representation of the
:ref:`cov_reference_table-name`.
If it `None``, the :ref:`com_cov_reference-name` routine
is used to create the cov_reference table.

{xrst_end create_all_node_db}
'''
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# is_descendant
def is_descendant(node_table, ancestor_node_id, this_node_id) :
   if this_node_id == ancestor_node_id :
      return True
   while this_node_id is not None :
      this_node_id = node_table[this_node_id]['parent']
      if this_node_id == ancestor_node_id :
         return True
   return False
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.create_all_node_db
def create_all_node_db(
   all_node_database         = None,
   option_all                = None,
   fit_goal_table            = None,
   split_reference_table     = None,
   node_split_table          = None,
   mulcov_freeze_table       = None,
   omega_grid                = None,
   omega_data                = None,
   cov_reference_table       = None,
) :
   if fit_goal_table is None :
      fit_goal_table = list()
   if split_reference_table is None :
      split_reference_table = list()
   if node_split_table is None :
      node_split_table = list()
   if mulcov_freeze_table is None :
      mulcov_freeze_table = list()
   #
   assert type(all_node_database)      == str
   assert type(option_all)             == dict
   assert type(fit_goal_table)         == list
   assert type(split_reference_table)  == list
   assert type(node_split_table)       == list
   assert type(mulcov_freeze_table)    == list
   assert type(cov_reference_table)    == list or cov_reference_table == None
   # END_DEF
   #
   # some asserts
   if omega_grid is None :
      assert omega_data is None
   else :
      assert type(omega_grid) is dict
      assert type(omega_data) is dict
   #
   assert 'root_database'      in option_all
   assert 'root_node_name'     in option_all
   assert 'result_dir'         in option_all
   #
   # n_split
   n_split = 1
   if len(split_reference_table) > 0 :
      n_split = len(split_reference_table)
   #
   # -------------------------------------------------------------------------
   # Read root node database
   # -------------------------------------------------------------------------
   # root_connection
   new                = False
   root_database      = option_all['root_database']
   root_connection    = dismod_at.create_connection(root_database, new)
   #
   # age_table
   tbl_name  = 'age'
   age_table = dismod_at.get_table_dict(root_connection, tbl_name)
   #
   # time_table
   tbl_name   = 'time'
   time_table = dismod_at.get_table_dict(root_connection, tbl_name)
   #
   # covariate_table
   tbl_name        = 'covariate'
   covariate_table = dismod_at.get_table_dict(root_connection, tbl_name)
   #
   # node table
   tbl_name   = 'node'
   node_table = dismod_at.get_table_dict(root_connection, tbl_name)
   #
   # option table
   tbl_name     = 'option'
   option_table = dismod_at.get_table_dict(root_connection, tbl_name)
   #
   # data_table
   if cov_reference_table == None :
      tbl_name   = 'data'
      data_table = dismod_at.get_table_dict(root_connection, tbl_name)

   #
   # root_connection
   root_connection.close()
   #
   # node_name2id
   node_name2id = dict()
   for (node_id, row) in enumerate(node_table) :
      node_name2id[ row['node_name'] ] = node_id
   #
   # root_node_id
   root_node_id = None
   for row in option_table :
      if row['option_name'] == 'parent_node_id' :
         root_node_id = row['option_value']
      elif row['option_name'] == 'parent_node_name' :
         root_node_id = node_name2id[ row['option_value'] ]
   assert root_node_id is not None
   #
   # root_node_name
   root_node_name = node_table[root_node_id]['node_name']
   if root_node_name != option_all['root_node_name' ] :
      tmp  = option_all['root_node_name']
      msg  = f'root_node_name in option_all is {tmp} '
      msg += f'while in the option table it is {root_node_name}'
      assert False, msg
   #
   # rel_covariate_id_set
   rel_covariate_id_set = set( range(len(covariate_table)) )
   if 'split_covariate_name' in option_all :
      split_covariate_name = option_all['split_covariate_name']
      split_covariate_id   = at_cascade.table_name2id(
         covariate_table, 'covariate', split_covariate_name
      )
      rel_covariate_id_set.remove(split_covariate_id)
   if 'absolute_covariates' in option_all :
      temp_list = option_all['absolute_covariates'].split()
      abs_covariate_id_set = set()
      for covariate_name in temp_list :
         covariate_id   = at_cascade.table_name2id(
            covariate_table, 'covariate', covariate_name
         )
      rel_covariate_id_set.remove(covariate_id)
   #
   # option_all_table
   option_all_table = list()
   for key in option_all :
      value = option_all[key]
      row   = { 'option_name' : key , 'option_value' : value }
      option_all_table.append( row )
   #
   # cov_reference_table
   if cov_reference_table == None :
      cov_reference_table = list()
      for node_id in range( len(node_table) ) :
         ancestor      = node_id
         while ancestor != root_node_id and ancestor != None :
            ancestor = node_table[ancestor]['parent']
         if ancestor == root_node_id :
            if len(split_reference_table) == 0 :
               split_reference_list = [ None ]
            else :
               split_reference_list = range( len(split_reference_table) )
            for split_reference_id in split_reference_list :
               reference_list = at_cascade.com_cov_reference(
                  option_all_table      = option_all_table,
                  split_reference_table = split_reference_table,
                  node_table            = node_table,
                  covariate_table       = covariate_table,
                  shift_node_id         = node_id,
                  split_reference_id    = split_reference_id,
                  data_table            = data_table,
               )
               for (covariate_id, reference) in enumerate(reference_list) :
                  row = {
                     'node_id'            : node_id ,
                     'split_reference_id' : split_reference_id,
                     'covariate_id'       : covariate_id,
                     'reference_value'    : reference_list[covariate_id],
                  }
                  cov_reference_table.append(row)
   # -------------------------------------------------------------------------
   # Write all node database
   # -------------------------------------------------------------------------
   # all_connection
   new             = True
   all_connection  = dismod_at.create_connection(all_node_database, new)
   #
   # cov_reference table
   tbl_name  = 'cov_reference'
   col_name  = [ 'node_id', 'split_reference_id', 'covariate_id' ]
   col_name += [ 'reference_value' ]
   col_type  = [ 'integer',            'integer',      'integer', 'real' ]
   row_list  = list()
   for row_table in cov_reference_table :
      row    = list()
      for key in col_name :
         row.append( row_table[key] )
      row_list.append( row )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # omega_age_grid table
   tbl_name    = 'omega_age_grid'
   col_name    = [ 'age_id'  ]
   col_type    = [ 'integer' ]
   row_list    = list()
   n_omega_age = None
   if not omega_grid is None :
      n_omega_age = len( omega_grid['age'] )
      for age_id in omega_grid['age'] :
         assert age_id < len(age_table)
         assert 0 <= age_id
         row_list.append( [ age_id ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # omega_time_grid table
   tbl_name     = 'omega_time_grid'
   col_name     = [ 'time_id'  ]
   col_type     = [ 'integer' ]
   row_list     = list()
   n_omega_time = None
   if not omega_grid is None :
      n_omega_time = len( omega_grid['time'] )
      for time_id in omega_grid['time'] :
         assert time_id < len(time_table)
         assert 0 <= time_id
         row_list.append( [ time_id ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # omega_all table
   tbl_name  = 'omega_all'
   col_name  = [ 'omega_all_value' ]
   col_type  = [  'real' ]
   row_list  = list()
   if not omega_data is None :
      node_list = omega_data.keys()
      for node_name in node_list :
         node_id = at_cascade.table_name2id(node_table, 'node', node_name)
         assert n_split == len( omega_data[node_name] )
         for k in range(n_split) :
            omega_list = omega_data[node_name][k]
            assert len(omega_list) == n_omega_age * n_omega_time
            for i in range(n_omega_age) :
               for j in range(n_omega_time) :
                  value   = omega_list[ i * n_omega_time + j]
                  row     = [ value ]
                  row_list.append( row )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # omega_index table
   tbl_name  = 'omega_index'
   col_name  = [ 'node_id', 'split_reference_id', 'omega_all_id' ]
   col_type  = [ 'integer', 'integer',             'integer' ]
   row_list  = list()
   if not omega_data is None :
      omega_all_id = 0
      for node_name in node_list :
         node_id = at_cascade.table_name2id(node_table, 'node', node_name)
         for k in range(n_split) :
            if len(split_reference_table) == 0 :
               split_reference_id = None
            else :
               split_reference_id = k
            row     = [ node_id, split_reference_id, omega_all_id ]
            row_list.append( row )
            omega_all_id += n_omega_age * n_omega_time
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # option_all table
   tbl_name = 'option_all'
   col_name = [ 'option_name', 'option_value' ]
   col_type = [ 'text',        'text']
   row_list = list()
   for key in option_all :
      option_name = key
      option_value = str( option_all[key] )
      row_list.append( [ option_name, option_value ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # split_reference table
   tbl_name = 'split_reference'
   col_name = [ 'split_reference_name', 'split_reference_value' ]
   col_type = [ 'text',                 'real']
   row_list = list()
   for row in split_reference_table :
      name  = row['split_reference_name']
      value = row['split_reference_value']
      row_list.append( [ name, value ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # fit_goal table
   tbl_name = 'fit_goal'
   col_name = [ 'node_id' ]
   col_type = [ 'integer' ]
   row_list = list()
   for row in fit_goal_table :
      if 'node_id' in row :
         node_id = row['node_id']
      else :
         node_name = row['node_name']
         node_id   = at_cascade.table_name2id(node_table, 'node', node_name)
      row_list.append( [ node_id ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # node_split table
   tbl_name = 'node_split'
   col_name = [ 'node_id' ]
   col_type = [ 'integer' ]
   row_list = list()
   for row in node_split_table :
      if 'node_id' in row :
         node_id = row['node_id']
      else :
         node_name = row['node_name']
         node_id   = at_cascade.table_name2id(node_table, 'node', node_name)
      row_list.append( [ node_id ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # mulcov_freeze table
   tbl_name = 'mulcov_freeze'
   col_name = [ 'fit_node_id', 'split_reference_id', 'mulcov_id' ]
   col_type = [ 'integer',     'integer',            'integer'   ]
   row_list = list()
   for row in mulcov_freeze_table :
      #
      # fit_node_id
      if 'fit_node_id' in row :
         fit_node_id = row['fit_node_id']
      else :
         fit_node_name = row['fit_node_name']
         fit_node_id   = at_cascade.table_name2id(
            node_table, 'node', fit_node_name
         )
      #
      # split_reference_id
      if 'split_reference_id' in row :
         split_reference_id = row['split_reference_id']
      else :
         split_reference_name = row['split_reference_name']
         if split_reference_name is None :
            split_reference_id = None
         else :
            split_reference_id   = at_cascade.table_name2id(
               split_reference_table,
               'split_reference',
               split_reference_name
         )
      #
      # mulcov_id
      mulcov_id = row['mulcov_id']
      #
      # row_list
      row_list.append( [ fit_node_id, split_reference_id, mulcov_id ] )
   dismod_at.create_table(
      all_connection, tbl_name, col_name, col_type, row_list
   )
   #
   # close
   all_connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2024 Bradley M. Bell
r'''
{xrst_begin data_include}

Rows of Data Table That are Include for a Fit
#############################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

fit_database
************
This is the database for the job that we will be fitting.
All of the dismod_at hold out commands that will be used for the fit
must be executed; i.e., the hold_out column in the data_subset table
is the same as will be used for the fit.

root_database
*************
This is the root node database. It contains the dismod_at tables
that are the same for all the fits in the cascade.

data_include_table
******************
This is the rows of the data table that are included in the fit; i.e.,
the rows that satisfy the following conditions:

#. The node is the node being fit or a descendant of the node being fit.
#. All the covariates are within their maximum difference limits.
#. The hold_out value in the data table is zero
#. The corresponding hold_out value in the data_subset table is zero.
#. The corresponding integrand is not in the option table hold_out list.

{xrst_end data_include}
'''
import at_cascade
#
# BEGIN_DEF
# at_cascade.data_include
def data_include(
   fit_database,
   root_database
) :
   assert type( fit_database ) == str
   assert type( root_database ) == str
   # END_DEF
   #
   # fit_or_root
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   #
   # data_subset_table
   data_subset_table = fit_or_root.get_table('data_subset')
   #
   # data_table
   data_table = fit_or_root.get_table('data')
   #
   # integrand_table
   integrand_table = fit_or_root.get_table('integrand')
   #
   # option_table
   option_table = fit_or_root.get_table('option')
   #
   # fit_or_root
   fit_or_root.close()
   #
   # hold_out_name_list
   hold_out_integrand = None
   for row in option_table :
      if row['option_name'] == 'hold_out_integrand' :
         hold_out_integrand = row[ 'option_value' ]
   if hold_out_integrand == None :
      hold_out_name_list = list()
   else :
      hold_out_name_list = hold_out_integrand.split()
   #
   # hold_out_id_list
   hold_out_id_set   = set()
   for integrand_name in hold_out_name_list :
      integrand_id = at_cascade.table_name2id(
         integrand_table, 'integrand', integrand_name
      )
      hold_out_id_set.add(integrand_id)
   #
   # data_include_table
   data_include_table = list()
   for subset_row in data_subset_table :
      #
      # data_id
      data_id = subset_row['data_id']
      #
      # data_row
      data_row = data_table[data_id]
      #
      # integrand_id
      integrand_id = data_row['integrand_id']
      #
      # hold_out
      hold_out = integrand_id in hold_out_id_set
      hold_out = hold_out or bool( data_row['hold_out'] )
      hold_out = hold_out or bool( subset_row['hold_out'] )
      #
      # data_include_table
      if not hold_out :
         data_include_table.append(data_row)
   #
   # BEGIN_RETURN
   assert type( data_include_table ) == list
   if len( data_include_table ) > 0 :
      assert type( data_include_table[0] ) == dict
      assert data_include_table[0]['hold_out'] == 0
   return data_include_table
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import csv
"""
{xrst_begin csv.read_table}

Create A Table from a CSV File
##############################

Prototype
*********
{xrst_literal ,
   BEGIN_DEF, END_DEF
   BEGIN_RETURN, END_RETURN
}


file_name
*********
is a ``str`` with the name of the CSV file.
The first line of the file is the header line and the others contain
the data.
The header value in the j-th column of the first line is the
corresponding column name.

table
*****
the return value *table* is a ``list`` of ``dict``.
We use *n* to denote the length of the list which is
the number of lines in the file minus one.
For *i*, an ``int`` ,  between zero and *n* -1,
and each column name *key* , a string,
*table* [ *i* ][ *key* ] is the ``str`` in line *i* +2 and column *key*.

Example
*******
:ref:`csv.table-name`

{xrst_end csv.read_table}
"""
# BEGIN_DEF
# at_cascade.csv.read_table
def read_table(file_name) :
   assert type(file_name)  == str
   # END_DEF
   #
   file_ptr   = open(file_name)
   reader     = csv.DictReader(file_ptr)
   table      = list()
   for row in reader :
      table.append(row)
   file_ptr.close()
   #
   # BEGIN_RETURN
   # ...
   assert type(table) == list
   return table
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin csv.pre_parallel}

Predict With Specified Maximum Number of Processes
##################################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

fit_dir
*******
Same as the csv fit :ref:`csv.fit@fit_dir` .

sim_dir
*******
Same as :ref:`csv.predict@sim_dir` .


all_covariate_table
*******************
This is an in memory representation of
:ref:`csv.fit@Input Files@covariate.csv` .

fit_goal_set
************
This set contains the node that we are required to fit; i.e.,
the nodes in :ref:`csv.fit@Input Files@fit_goal.csv` .
Ancestors between these nodes and the root node are also fit.

start_job_name
**************
Is the name of the job (fit) that the predictions should start at.
This is a node name, followed by a period, followed by a sex.
Only this fit, and its descendants, will be included in the predictions.
If this argument is None, all of the jobs (fits) will be included.

max_job_depth
*************
This is the number of generations below start_job_name that are included
in the predictions.
If max_job_depth is zero,  only the start job will be included.
If max_job_depth is None,  start job and all its descendants are included.

option_predict
**************
This is an in memory representation of
:ref:`csv.predict@Input Files@option_predict.csv` .

{xrst_end csv.pre_parallel}
r'''
# ----------------------------------------------------------------------------
import at_cascade
import dismod_at
import multiprocessing
import numpy
# ----------------------------------------------------------------------------
# shared_memory_prefix = get_shared_memory_prefix(all_node_database)
def get_shared_memory_prefix(all_node_database) :
   assert type(all_node_database) == str
   #
   connection           = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table     = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   shared_memory_prefix = ""
   for row in option_all_table :
      if row['option_name'] == 'shared_memory_prefix' :
         shared_memory_prefix = row['option_value']
   return shared_memory_prefix
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.csv.pre_parallel
def pre_parallel(
   fit_dir,
   sim_dir,
   all_covariate_table,
   fit_goal_set,
   start_job_name,
   max_job_depth,
   option_predict,
) :
   assert type(fit_dir)                     == str
   assert sim_dir == None or type(sim_dir)  == str
   assert type(all_covariate_table)         == list
   assert type( all_covariate_table[0] )    == dict
   assert type(fit_goal_set)                == set
   assert type( next(iter(fit_goal_set) ))  == str
   assert type( option_predict )            == dict
   # END_DEF
   # ----------------------------------------------------------------------
   # job_status_name
   job_status_name = [
      'skip' , # will not process this job
      'ready', # job is readiy to run
      'run'  , # job is running
      'done' , # job finished running
   ]
   job_status_skip  = job_status_name.index( 'skip' )
   job_status_ready = job_status_name.index( 'ready' )
   job_status_run   = job_status_name.index( 'run' )
   job_status_done  = job_status_name.index( 'done' )
   # -------------------------------------------------------------------------
   #
   # max_number_cpu
   max_number_cpu = option_predict['max_number_cpu']
   #
   # all_node_db
   all_node_db = f'{fit_dir}/all_node.db'
   #
   # root_database
   root_database      = f'{fit_dir}/root.db'
   #
   # root_node_name
   root_node_name = at_cascade.get_parent_node(root_database)
   #
   # node_table, covariate_table
   connection      = dismod_at.create_connection(
      root_database, new = False, readonly = True
   )
   node_table      = dismod_at.get_table_dict(connection, 'node')
   covariate_table = dismod_at.get_table_dict(connection, 'covariate')
   connection.close()
   #
   # root_split_reference_id
   root_split_reference_id = None
   sex_value               = None
   for row in covariate_table :
      if row['covariate_name'] == 'sex' :
            sex_value = row['reference']
   for (row_id, row) in enumerate( at_cascade.csv.split_reference_table ) :
      if row['split_reference_value'] == sex_value :
         root_split_reference_id = row_id
   assert root_split_reference_id != None
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(
      node_table, 'node', root_node_name
   )
   #
   # split_reference_table
   connection      = dismod_at.create_connection(
      all_node_db, new = False, readonly = True
   )
   split_reference_table = \
      dismod_at.get_table_dict(connection, 'split_reference')
   connection.close()
   assert split_reference_table == at_cascade.csv.split_reference_table
   #
   # job_table
   job_table = at_cascade.create_job_table(
      all_node_database          = all_node_db              ,
      node_table                 = node_table               ,
      start_node_id              = root_node_id             ,
      start_split_reference_id   = root_split_reference_id  ,
      fit_goal_set               = fit_goal_set             ,
   )
   #
   # start_job_id
   if start_job_name == None :
      start_job_id = 0
   else :
      start_job_id = None
   for (job_id, row) in enumerate(job_table) :
      if row['job_name'] == start_job_name :
         start_job_id = job_id
   if start_job_id == None :
      root_job_name = job_table[0]['job_name']
      msg  = f'start_job_name = {start_job_name} is not a valid job name '
      msg += 'for this cascade.\n'
      msg += f'The root job name for this cascade is {root_job_name}.'
      assert False, msg
   #
   # at_cascade_log_dict
   # get log for all jobs so that we can find ancestor jobs that completed
   log_start_job_id = 0
   if max_job_depth == None :
      log_max_job_depth = None
   elif start_job_id == None :
      log_max_job_depth = max_job_depts
   else :
      log_max_job_depth = max_job_depth + start_job_id
   at_cascade_log_dict = at_cascade.check_log(
      message_type       = 'at_cascade'         ,
      all_node_database  = all_node_db          ,
      root_database      = root_database   ,
      job_table          = job_table            ,
      start_job_id       = log_start_job_id     ,
      max_job_depth      = log_max_job_depth    ,
   )
   #
   # n_job
   n_job = len( job_table )
   assert job_table[0]['parent_job_id'] == None
   #
   # process_list
   process_list = list()
   #
   # predict_job_id_list
   predict_job_id_list    = list()
   for predict_job_id in range(n_job) :
      #
      # include_this_job
      job_depth = at_cascade.job_descendant(
         job_table, start_job_id, predict_job_id
      )
      include_this_job = False
      if job_depth != None :
         if max_job_depth == None :
            include_this_job = True
         else :
            include_this_job = job_depth <= max_job_depth
      if include_this_job :
         predict_job_id_list.append( predict_job_id )
   #
   # n_predict
   n_predict = len(predict_job_id_list)
   #
   # ----------------------------------------------------------------------
   # shared_job_status_name
   shared_memory_prefix      = get_shared_memory_prefix(all_node_db)
   start_name                = job_table[start_job_id]['job_name']
   shared_memory_prefix_plus = shared_memory_prefix + f'_pre_{start_name}'
   shared_job_status_name    = shared_memory_prefix_plus + '_job_status'
   print(f'create: {shared_job_status_name} shared memory')
   #
   # shm_job_status, shared_job_status
   tmp    = numpy.empty(len(job_table), dtype = int )
   mapped = at_cascade.map_shared( shared_job_status_name )
   shm_job_status = multiprocessing.shared_memory.SharedMemory(
      create = True, size = tmp.nbytes, name = mapped
   )
   shared_job_status = numpy.ndarray(
      tmp.shape, dtype = tmp.dtype, buffer = shm_job_status.buf
   )
   shared_job_status[:] = job_status_skip
   for predict_job_id in predict_job_id_list :
      shared_job_status[predict_job_id] = job_status_ready
   #
   # shared_lock
   shared_lock = multiprocessing.Lock()
   #
   # -------------------------------------------------------------------------
   #
   # process_list
   # execute pre_one_process for each process in process_list
   n_spawn      = min(n_predict - 1, max_number_cpu - 1)
   print( f'Predict: n_predict = {n_predict}, n_spawn = {n_spawn}' )
   process_list = list()
   for i in range(n_spawn) :
      p = multiprocessing.Process(
         target = at_cascade.csv.pre_one_process,
         args=(
            fit_dir,
            sim_dir,
            option_predict,
            all_node_db,
            all_covariate_table,
            job_table,
            node_table,
            root_node_id,
            root_split_reference_id,
            at_cascade_log_dict,
            job_status_name,
            shared_job_status_name,
            shared_lock,
         )
      )
      p.start()
      process_list.append(p)
   #
   # pre_one_process
   # use this process as well to execute pre_one_process
   at_cascade.csv.pre_one_process(
      fit_dir,
      sim_dir,
      option_predict,
      all_node_db,
      all_covariate_table,
      job_table,
      node_table,
      root_node_id,
      root_split_reference_id,
      at_cascade_log_dict,
      job_status_name,
      shared_job_status_name,
      shared_lock,
   )
   #
   # join
   # wait for all the processes to finish
   for p in process_list :
      p.join()
   #
   # pre_user
   at_cascade.csv.pre_user(
      fit_dir,
      sim_dir,
      job_table,
      start_job_name,
      predict_job_id_list,
      node_table,
      root_node_id,
      root_split_reference_id,
      root_database,
   )
   #
   # shm_job_status
   print(f'remove: {shared_job_status_name} shared memory')
   shm_job_status.close()
   shm_job_status.unlink()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import csv
r"""
{xrst_begin csv.write_table}
{xrst_spell
  len
}

Create A CSV File from a Table
##############################

Prototype
*********
{xrst_literal
   BEGIN_DEF
   END_DEF
}

table
*****
This must be  a ``list`` of ``dict``.

columns
*******
This is a ``list`` of ``str`` specifying the keys in the table dictionary
that are written to the file.
If this argument is ``None`` ,
*table* [0].keys() is used as its default value.
If :ref:`csv.read_table-name` returns am empty table,
:ref:`csv.get_header-name` can be used to determine the column names.

file_name
*********
is a ``str`` with the name of the CSV file.
Upon return, this file has ``len(`` *table* ``)`` + 1 lines,
``len(`` *columns* ``)`` columns, with the following values

.. list-table::

   * - columns[0]
     - columns[1]
     - columns[2]
     - ...
   * - table[0][ columns[0] ]
     - table[0][ columns[1] ]
     - table[0][ columns[2] ]
     - ...
   * - table[1][ columns[0] ]
     - table[1][ columns[1] ]
     - table[1][ columns[2] ]
     - ...
   * - \:
     - \:
     - \:
     - ...

Example
*******
:ref:`csv.table-name`


{xrst_end csv.write_table}
"""
# BEGIN_DEF
# at_cascade.csv.write_table
def write_table(
   file_name  = None,
   table      = None,
   columns    = None,
) :
   assert type(file_name)  == str
   assert type(table)      == list
   assert type(columns) == list or columns == None
   if columns == None :
      columns = table[0].keys()
   # END_DEF
   #
   file_ptr    = open(file_name, 'w')
   writer      = csv.DictWriter(file_ptr, fieldnames = columns)
   writer.writeheader()
   writer.writerows( table )
   file_ptr.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin continue_cascade}

Continue Cascade From a Fit Node
################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
Sometimes when running the cascade, the fit or statistics for a node fails.
This may be because of something that happened on the system,
or because of some of the settings in the :ref:`glossary@root_database`.
This routine enables you to continue the cascade from such a node.

all_node_database
*****************
is a python string specifying the location of the
:ref:`all_node_db-name`
relative to the current working directory.
This argument can't be ``None``.

fit_database
************
is a python string specifying the location of a dismod_at database
relative to the current working directory.
This is a :ref:`glossary@fit_database` with the
final state after a run of
:ref:`cascade_root_node-name` or :ref:`continue_cascade-name`
that includes fitting this database.
The *fit_database* is not changed, it is only used
to identify which child jobs to fit.

fit_goal_set
************
This is a ``set`` with elements of type ``int`` (``str``)
specifying the node_id (node_name) for each element of the
:ref:`glossary@fit_goal_set` .
Each such node must be the root node, or a descendant of the root node.
In addition, it must be in the :ref:`fit_goal_table-name` ,
or an ancestor of a node in the fit goal table.

fit_type_list
*************
This is a list with one or two elements
and its possible elements are ``both`` and ``fixed``.
For each job, the first type of fit is attempted.
If it fails, and there is a second type of fit, it is attempted.
If it also fails, the corresponding job fails.

shared_unique
*************
Under normal circumstances, you should use the empty string (default value)
for this parameter.

#. Parallel runs of continue_cascade must use different values of
   *shared_unique* so the corresponding shared memory names are different;
   see :ref:`fit_parallel@shared_unique` .

#. The *shared_unique* special case is where you are running (in parallel)
   continue_cascade with the same *fit_database*, and disjoint
   *fit_goal_set* . (The intersection of disjoint sets is empty.)

#. In the special case above, the suggested value for *shared_unique*
   is ``_`` *node_name* , where *node_name* is the name of one of the nodes
   in the *fit_goal_set*. It may be necessary to include the value
   of the splitting covariate in *shared_unique*  .
   (The splitting covariate is sex in the :ref:`csv.fit-name` case.)

{xrst_end   continue_cascade}
'''
import time
import os
import multiprocessing
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.continue_cascade
def continue_cascade(
   all_node_database = None,
   fit_database      = None,
   fit_goal_set      = None,
   fit_type_list     = [ 'both', 'fixed' ],
   shared_unique     = '',
) :
   assert type(all_node_database) == str
   assert type(fit_database) == str
   assert type(fit_goal_set)      == set
   assert type(fit_type_list)     == list
   assert type(shared_unique)     == str
   # END_DEF
   #
   # split_reference_table, option_all, node_split_table, fit_goal
   connection       = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   node_split_table = dismod_at.get_table_dict(connection, 'node_split')
   split_reference_table = \
      dismod_at.get_table_dict(connection, 'split_reference')
   connection.close()
   #
   # result_dir, root_node_name, root_database,
   # max_number_cpu, refit_split
   result_dir         = None
   root_node_name     = None
   root_database      = None
   max_number_cpu     = 1
   refit_split        = False
   for row in option_all_table :
      if row['option_name'] == 'result_dir' :
         result_dir = row['option_value']
      if row['option_name'] == 'root_node_name' :
         root_node_name = row['option_value']
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
      if row['option_name'] == 'max_number_cpu' :
         max_number_cpu = int( row['option_value'] )
      if row['option_name'] == 'refit_split' :
         refit_split = row['option_value'] == 'true'
   assert result_dir is not None
   assert root_node_name is not None
   assert root_database is not None
   #
   # node_table, covariate_table, fit_integrand
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   node_table      = fit_or_root.get_table('node')
   covariate_table = fit_or_root.get_table('covariate')
   fit_integrand   = at_cascade.get_fit_integrand(fit_or_root)
   fit_or_root.close()
   #
   # root_node_id
   root_node_id = at_cascade.table_name2id(node_table, 'node', root_node_name)
   #
   # root_split_reference_id
   root_split_reference_id = None
   for row in option_all_table :
      if row['option_name'] == 'root_split_reference_name' :
         root_split_reference_name = row['option_value']
         root_split_reference_id = at_cascade.table_name2id(
            split_reference_table,
            'split_reference',
            root_split_reference_name
         )
   #
   # fit_node_id
   fit_node_name = at_cascade.get_parent_node(fit_database)
   fit_node_id   = at_cascade.table_name2id(node_table, 'node', fit_node_name)
   #
   # fit_split_reference_id
   if len(split_reference_table) == 0 :
      fit_split_reference_id = None
   else :
      cov_info = at_cascade.get_cov_info(
         option_all_table, covariate_table, split_reference_table
      )
      fit_split_reference_id = cov_info['split_reference_id']
   #
   # job_table
   job_table = at_cascade.create_job_table(
      all_node_database          = all_node_database,
      node_table                 = node_table,
      start_node_id              = fit_node_id,
      start_split_reference_id   = fit_split_reference_id,
      fit_goal_set               = fit_goal_set,
   )
   #
   # check job_table[0]
   assert fit_node_id == job_table[0]['fit_node_id']
   assert fit_split_reference_id == job_table[0]['split_reference_id']
   #
   # start_job_id
   start_job_id = 0
   #
   # lock
   lock = multiprocessing.Lock()
   #
   # skip_start_job
   skip_start_job = True
   #
   # fit_one_process
   at_cascade.fit_parallel(
      job_table         = job_table ,
      start_job_id      = start_job_id,
      all_node_database = all_node_database,
      node_table        = node_table,
      fit_integrand     = fit_integrand,
      skip_start_job    = skip_start_job,
      max_number_cpu    = max_number_cpu,
      fit_type_list     = fit_type_list,
      shared_unique     = shared_unique,
   )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import multiprocessing
import queue
import dismod_at
import at_cascade
import copy
import os
import time
r'''
{xrst_begin csv.predict}
{xrst_spell
  avg
  avgint
  cpus
  meas
  pdf
  tru
  std
}

Prediction for a CSV Fit
########################

Prototype
*********
{xrst_literal
   # BEGIN_PREDICT
   # END_PREDICT
}

Example
*******
:ref:`csv.sim_fit_pred-name` .

fit_dir
*******
Same as the csv fit :ref:`csv.fit@fit_dir` .

sim_dir
*******
If this is None, the file
:ref:`csv.predict@Output Files@tru_predict.csv` is not created.
Otherwise, :ref:`csv.simulate@sim_dir` is the directory
used to simulate the data for this fit and the file
tru_predict.csv is created.

start_job_name
**************
Is the name of the job (fit) that the predictions should start at.
This is a node name, followed by a period, followed by a sex.
Only this fit, and its descendants, will be included in the predictions.
If this argument is None, all of the jobs (fits) will be included.

max_job_depth
*************
This is the number of generations below start_job_name that are included;
see :ref:`job_descendant@Node Depth Versus Job Depth`
and note that sex is the :ref:`option_all_table@split_covariate_name` .
If max_job_depth is zero,  only the start job will be included.
If max_job_depth is None,  start job and all its descendants are included;

Input Files
***********

option_predict.csv
==================
This csv file has two columns,
one called ``name`` and the other called ``value``.
The rows of this table are documented below by the name column.
If an option name does not appear, or the corresponding value is empty,
the default value is used for the option.
The final value for each of the options is reported in the file
:ref:`csv.predict@Output Files@option_predict_out.csv` .
Because each option has a default value,
new option are added in such a way that
previous option_predict.csv files are still valid.

db2csv
------
If this boolean option is true,
the dismod_at `db2csv_command`_ is used to generate the csv files
corresponding to each :ref:`csv.fit@Output Files@dismod.db` .
This is only done for (node, sex) pairs that have samples; i.e.,
a successful fit and posterior samples.
If this option is true, the csv files will make it more difficult
to see the tree structure corresponding to the ``dismod.db`` files.
The default value for this option is false .

.. _db2csv_command: https://dismod-at.readthedocs.io/latest/db2csv_command.html

descendant_std_factor
---------------------
This factor scales an ancestor fit posterior samples before predicting
for a descendant job; i.e., (node, sex) pair.
It must be greater than zero and it's default value is 1.
It is only used when predicting for a job that does **not** have samples.
In this case the closest ancestor that does have samples
is used to predict for the (node, sex) pair; see
:ref:`csv.ancestor_fit-name`.
For an example, see :ref:`csv.predict_descend-name` .

float_precision
---------------
This integer is the number of decimal digits of precision to
include for float values in the output csv files.
The default value for this option is 5.

.. _plot_curve: https://dismod-at.readthedocs.io/latest/plot_curve.html

max_number_cpu
--------------
This integer is the maximum number of cpus (processes) to use
This must be greater than zero. If it is one, the jobs are run
sequentially, more output is printed to the screen, and the program
can be cleanly stopped with a control-C.
The default value for this option is
{xrst_code py}
   max_number_cpu = max(1, multiprocessing.cpu_count() - 1)
{xrst_code}

plot
----
The default value for this option is false .
If this boolean option is true,
a ``data_plot.pdf`` and ``rate_plot.pdf`` file is created for each
:ref:`csv.fit@Output Files@dismod.db` database.
This is only done for (node, sex) pairs that have samples; i.e.,
a successful fit and posterior samples.
The data plot includes a maximum of 1,000 randomly chosen points for each
integrand in the predict_integrand.csv file.
The rate plot includes all the non-zero rates.
These are no effect rates; i.e., they are the estimated rate
for this node and sex without any covariate effects.
Predictions with covariate effects can be found in the csv
:ref:`csv.predict@Output Files` .

zero_meas_value
---------------
If this boolean option is true, the
:ref:`csv.fit@Input Files@mulcov.csv@type@meas_value` covariate
multipliers are set to zero during the predictions
(instead of their simulation values, fit, or sample values).
The default value for this option is false .

number_sample_predict
---------------------
This integer option specifies the number of samples generated for each
prediction.  Its default value is the value of
:ref:`csv.fit@Input Files@option_fit.csv@number_sample` in
:ref:`csv.fit@Input Files@option_fit.csv`.
If number_sample_predict does not appear in option_predict.csv,
and number_sample does not appear in option_fit.csv,
the default value for number_sample
is the value used for number_sample_predict.

covariate.csv
=============
Same as the csv fit
:ref:`csv.fit@Input Files@covariate.csv` .

fit_goal.csv
============
Same as the csv fit
:ref:`csv.fit@Input Files@fit_goal.csv` .

option_fit.csv
==============
The value option_fit.csv
:ref:`csv.fit@Input Files@option_fit.csv@refit_split` value is used.

predict_integrand.csv
=====================
This is the list of integrands at which predictions are made
and stored in :ref:`csv.predict@Output Files@fit_predict.csv` .


{xrst_comment ---------------------------------------------------------------}


Output Files
************

option_predict_out.csv
======================
This is a copy of
:ref:`csv.predict@Input Files@option_predict.csv` with the default
filled in for missing values.

fit_predict.csv
===============
#. If :ref:`csv.predict@start_job_name` is None,
   ``fit_predict.csv`` contains the predictions for all the fits.
   These predictions for all of the nodes at the age, time and
   covariate values specified in covariate.csv.
   The prediction is done using the optimal variable values.

#. If :ref:`csv.predict@start_job_name` is not None,
   the predictions are only for jobs at or below the starting job.
   In addition, the predictions are stored below *fit_dir* in the file

      ``predict/fit_``\ *start_job_name*\ ``.csv``

   and not in ``fit_predict.csv`` .


avgint_id
---------
Each avgint_id corresponds to a different value for age, time,
or integrand in the fit_predict file.
The age and time values comes from the covariate.csv file.
The integrands values come from the predict_integrand.csv file
and the covariate multiplier list.

sample_index
------------
Each sample_index corresponds to an independent random sample
of the model variables.

#. If :ref:`option_all_table@sample_method` is asymptotic,
   model variables for each sample are Gaussian correlated with mean equal to
   the optimal value and variance equal to the asymptotic approximation.
#. If :ref:`option_all_table@sample_method` is censor_asymptotic,
   model variables are the same as for asymptotic expect that values above
   (below) their upper bound (lower bound) are converted to the corresponding
   bound.
#. If :ref:`option_all_table@sample_method` is simulate,
   the model variables for each sample at the optimal values corresponding
   to an independent data set.

integrand_name
--------------
is the integrand for this sample is equal to the integrand names
in predict_integrand.csv
The integrand names ``mulcov_0`` , ``mulcov_1`` , ...
corresponds to the first , second , ...
covariate multiplier in the csv fit
:ref:`csv.fit@Input Files@mulcov.csv` file.

avg_integrand
-------------
This float is the mode value for the average of the integrand,
with covariate and other effects but without measurement noise.

node_name
---------
is the node name for this sample is predicting for.
This cycles through all the nodes in covariate.csv.

sex
---
is the sex, female, both, or male, that the predictions are for.

fit_node_name
-------------
is the node name corresponding to the fit, and samples, that was used
to do these predictions.
This identifies the nearest ancestor that had a successful fit and samples.

fit_sex
-------
is the sex corresponding to the fit, and samples, that were used
to do these prediction.

posterior
.........
If *fit_node_name* and *fit_sex* are the same as *node_name* and *sex* ,
the fit and samples succeeded for this *node_name* and *sex* and
this row contains a posterior prediction for this *node_name* and *sex* .

prior
.....
If *fit_node_name* is not the same as *node_name* ,
or *fit_sex* is not the same as *sex* ,
this row contains a prior prediction for this *node_name* and *sex* .
The pair ( *fit_node_name* , *fit_sex* ) correspond to
the closest ancestor fit that was successful.

age
---
is the age for this prediction and is one of
the ages in covariate.csv.

time
----
is the time for this prediction and is one of
the times in covariate.csv.

covariate_names
---------------
The rest of the columns are covariate names and contain the value
of the corresponding covariate in
:ref:`csv.fit@Input Files@covariate.csv` .

tru_predict.csv
===============
If :ref:`csv.predict@sim_dir` is None, this file is not created.
Otherwise, this file contains the predictions for the model variables
corresponding to the simulation.
It is similar to :ref:`csv.predict@Output Files@fit_predict.csv`
with the following differences:

#. The first line (header line) is the same in this file and
   fit_predict.csv.
#. If the other lines, in both files, are sorted by
   ( *node_name* , *avgint_id* ) ,
   the other lines are the same except for the value in the
   avg_integrand column.
#. The model variables and true values, are for the
   *fit_node_name* and *fit_sex* . Hence this does not really represent truth
   unless these are the same as *node_name* and *sex* .

sam_predict.csv
===============
This is a sampling of the predictions,
using the posterior distribution of the model variables:
It is similar to :ref:`csv.predict@Output Files@fit_predict.csv`
with the following differences:

#. The first line (header line) is the same in this file and
   fit_predict.csv except that sam_predict.csv has an extra column
   named sample_index.
#. Suppose that the other lines in sam_predict.csv and fit_predict.csv
   are sorted by ( *node_name* , *avgint_id* ) .
#. Let *n_sample* be the number of other lines in sam_predict.csv divided by
   the number of other lines in fit_predict.csv.
#. For each line in fit_predict.csv (not counting the header line),
   there are *n_sample* lines in sam_predict.csv,
   that are the same as the line in fit_predict.csv except for the value in the
   avg_integrand column and the extra sample_index column in
   sam_predict.csv.

start_job_name
--------------
If *start_job_name* is not None,
the predictions are only for jobs at or below the starting job.
In addition, the predictions are stored below *fit_dir* in the file

   ``predict/sam_``\ *start_job_name*\ ``.csv``

and not in ``sam_predict.csv`` .

sample_index
------------
For each sample_index value, there is a complete set of all the values
in the fit_predict.csv table.
A different (independent) sample from of the model variables
from their posterior distribution is used to do the predictions for
each sample index.

{xrst_end csv.predict}
'''
# ----------------------------------------------------------------------------
# Returns the node indices for the ancestors of the node specified by node_id.
# The node specified by node_id is included in this set.
def ancestor_set(node_table, node_id) :
   result = { node_id }
   while node_table[node_id]['parent'] != None :
      node_id = node_table[node_id]['parent']
      result.add(node_id)
   return result
# ----------------------------------------------------------------------------
# Sets global global_option_value to dict representation of option_predict.csv
#
# fit_dir
# is the directory where the input csv files are located.
#
# option_table :
# is the list of dict corresponding to option_predict.csv
#
# number_sample_fit
# is equal to number_sample in the option_fit.csv file.
#
# top_node_name
# is the name of the top node in the node tree
#
# option_predict_out.csv
# As a side effect, this routine write a copy of the option table
# with the default values filled in.
#
# global_option_value[name] :
# is the option value corresponding the specified option name.
# Here name is a string and value
# has been converted to its corresponding type.
#
global_option_value = None
def set_global_option_value(
   fit_dir, option_table, number_sample_fit, top_node_name
) :
   global global_option_value
   assert type(global_option_value) == dict or global_option_value == None
   assert type(option_table) == list
   if len(option_table) > 0 :
      assert type( option_table[0] ) == dict
   assert type(number_sample_fit) == int
   assert type(top_node_name) == str
   #
   # user
   user = os.environ.get('USER')
   if user == None :
      user = 'none'
   else :
      user = user.replace(' ', '_')
   #
   # option_default
   max_number_cpu = max(1, multiprocessing.cpu_count() - 1)
   random_seed    = int( time.time() )
   # BEGIN_SORT_THIS_LINE_PLUS_2
   option_default  = {
      'db2csv'                : (bool,  False)              ,
      'descendant_std_factor' : (float,   1.0)              ,
      'float_precision'       : (int,   5)                  ,
      'max_number_cpu'        : (int,   max_number_cpu)     ,
      'number_sample_predict' : (int,   number_sample_fit)  ,
      'plot'                  : (bool,  False)              ,
      'zero_meas_value'       : (bool,  False)              ,
   }
   # END_SORT_THIS_LINE_MINUS_2
   #
   # global_option_value
   line_number      = 0
   global_option_value = dict()
   for row in option_table :
      line_number += 1
      name         = row['name']
      if name in global_option_value :
         msg  = f'csv.predict: Error: line {line_number}'
         msg += ' in option_predict.csv\n'
         msg += f'the name {name} appears twice in this table'
         assert False, msg
      if not name in option_default :
         msg  = f'csv.predict: Error: line {line_number}'
         msg += ' in option_predict.csv\n'
         msg += f'{name} is not a valid option name'
         assert False, msg
      (option_type, default) = option_default[name]
      value                  = row['value']
      if value.strip() != '' :
         if option_type == bool :
            if value not in [ 'true', 'false' ] :
               msg  = f'csv.predict: Error: line {line_number}'
               msg += ' in option_predict.csv\n'
               msg += f'The value for {name} is not true or false'
               assert False, msg
            global_option_value[name] = value == 'true'
         else :
            global_option_value[name] = option_type( value )
   #
   # global_option_value
   for name in option_default :
      if name not in global_option_value :
         (option_type, default) = option_default[name]
         global_option_value[name] = default
   #
   if global_option_value['descendant_std_factor']  <= 0.0 :
      assert False, "predict.csv: descemdant_std_factor <= 0"
   #
   # option_predict_out.csv
   table = list()
   for name in global_option_value :
      value = global_option_value[name]

      if type(value) == bool :
         if value :
            value = 'true'
         else :
            value = 'false'
      row = { 'name' : name , 'value' : value }
      table.append(row)
   file_name = f'{fit_dir}/option_predict_out.csv'
   at_cascade.csv.write_table(file_name, table)
   #
   assert type(global_option_value) == dict
# ----------------------------------------------------------------------------
# BEGIN_PREDICT
# at_cascade.csv.predict
def predict(fit_dir, sim_dir=None, start_job_name=None, max_job_depth=None) :
   assert type(fit_dir)  == str
   assert sim_dir        == None or type(sim_dir) == str
   assert start_job_name == None or type(start_job_name) == str
   assert max_job_depth  == None or type(max_job_depth) == int
   # END_PREDICT
   #
   # dismod_node_table, dismod_option_table
   database     = f'{fit_dir}/root.db'
   connection   = dismod_at.create_connection(
      database, new = False, readonly = True
   )
   dismod_node_table   = dismod_at.get_table_dict(connection, 'node')
   dismod_option_table = dismod_at.get_table_dict(connection, 'option')
   connection.close()
   #
   # top_node_name
   top_node_name = None
   for row in dismod_node_table :
      if row['parent'] == None :
         if top_node_name != None :
            msg = 'root.db: node table: more than one node has no parent'
            assert False, msg
         top_node_name = row['node_name']
   if top_node_name == None :
      msg = 'root.db: node_table: no node has None for parent'
      assert False, msg
   #
   # refit_split, number_sample_fit
   # 2DO: It would be better to get these values from the option_all table
   # (see where csv.fit writes the option_all table).
   refit_split       = True
   number_sample_fit = 20
   option_table      = at_cascade.csv.read_table(f'{fit_dir}/option_fit.csv')
   for row in option_table :
      if row['name'] == 'refit_split' :
         if row['value'] == 'false' :
            refit_split = False
      if row['name'] == 'number_sample':
         number_sample_fit = int(row['value'])
   #
   # global_option_value
   option_table = at_cascade.csv.read_table(f'{fit_dir}/option_predict.csv')
   set_global_option_value(
      fit_dir, option_table, number_sample_fit, top_node_name
   )
   #
   # start_node_name
   start_node_name = None
   if start_job_name == None :
      for row in dismod_option_table :
         if row['option_name'] == 'parent_node_name' :
            start_node_name = row['option_value']
         elif row['option_value'] == 'parent_node_id' :
            start_node_id   = int( row['option_value'] )
            start_node_name = dismod_node_table[parent_node_id]['node_name']
   else :
      index           = start_job_name.rindex('.')
      start_node_name = start_job_name[: index]
   assert start_node_name != None
   #
   # max_node_depth
   if max_job_depth == None :
      max_node_depth = None
   elif not refit_split :
      max_node_depth = max_job_depth
   elif start_job_name == None or start_job_name.endswith('.both') :
      if refit_split :
         if max_job_depth > 0 :
            max_node_depth = max_job_depth - 1
         else :
            max_node_depth = 0
      else :
         max_node_depth = max_job_depth
   else :
         max_node_depth = max_job_depth
   #
   # fit_goal_table
   file_name      = f'{fit_dir}/fit_goal.csv'
   fit_goal_table = at_cascade.csv.read_table(file_name)
   if len(fit_goal_table) == 0 :
      fit_goal_table = list()
      for node_id in range( len(dismod_node_table) ) :
         node_name = dismod_node_table[node_id]['node_name']
         row = { 'node_id' : node_id , 'node_name' : node_name}
         fit_goal_table.append( row )
   for row in fit_goal_table :
      node_name = row['node_name']
      node_id   = at_cascade.table_name2id(dismod_node_table, 'node', node_name)
      row['node_id'] = node_id
   #
   # fit_goal_set
   fit_goal_set   = set()
   start_node_id  = \
      at_cascade.table_name2id(dismod_node_table, 'node', start_node_name)
   for row in fit_goal_table :
      node_id   = row['node_id']
      node_list = [ node_id ]
      while node_id != start_node_id and node_id != None :
         node_id   = dismod_node_table[node_id]['parent']
         if node_id != None :
            node_list.append( node_id )
      if node_id == start_node_id :
         if max_node_depth == None :
            node_id   = node_list[0]
         elif len(node_list) <= max_node_depth + 1 :
            node_id   = node_list[0]
         else :
            node_id  = node_list[-max_node_depth - 1]
         node_name = dismod_node_table[node_id]['node_name']
         fit_goal_set.add( node_name )
   #
   if len(fit_goal_set) == 0 :
      msg  = f'Cannot find start_node_name = {start_node_name},\n'
      msg += 'or any of its children, in fit_goal.csv'
      assert False, msg
   #
   # covariate_table
   file_name       = f'{fit_dir}/covariate.csv'
   covariate_table = at_cascade.csv.read_table(file_name)
   #
   # predict
   at_cascade.csv.pre_parallel(
      fit_dir,
      sim_dir,
      covariate_table,
      fit_goal_set,
      start_job_name,
      max_job_depth,
      global_option_value,
   )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_fit_children}
{xrst_spell
  len
}

Determine the Set of Nodes to Fit
#################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

root_node_id
************
This is the node_id in the node_table for the
:ref:`glossary@root_node`.

fit_goal_set
************
This is a ``set`` with elements of type ``int``
specifying the node_id for each element of the
:ref:`glossary@fit_goal_set` or :ref:`fit_goal_table-name` .

node_table
**********
This is python list of python dictionaries
containing the dismod_at node table.

fit_children
************
The return value *fit_children* is a python list of python sets.

For each *node_id*
*fit_children* [ *node_id* ] is a the set of node ids that
are children of *node_id* and must be fit in order to fit
all the nodes in *fit_goal_set*.

Note that there must be at least one *node_id* in *fit_goal_set* ,
that is not an ancestor of any node in *fit_goal_set* and hence
len( *fit_children* [ *node_id* ] ) = 0 .

{xrst_end get_fit_children}
'''
# ----------------------------------------------------------------------------
import sys
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.get_fit_children
def get_fit_children(
   root_node_id  ,
   fit_goal_set  ,
   node_table    ,
) :
   assert type( root_node_id ) == int
   assert type( fit_goal_set ) == set
   assert type( node_table ) == list
   # END_DEF
   #
   # number of nodes
   n_node       = len( node_table )
   #
   # fit_children
   fit_children = list()
   for i in range(n_node) :
      fit_children.append( set() )
   #
   # goal_node_id
   for goal_node_id in fit_goal_set :
      assert type(goal_node_id) == int
      #
      # parend_id_list
      parent_id_list = list()
      parent_id  = node_table[goal_node_id]['parent']
      while parent_id != root_node_id and parent_id != None :
         parent_id_list.append( parent_id )
         parent_id  = node_table[parent_id]['parent']
      #
      # fit_children
      if parent_id == root_node_id :
         node_id = goal_node_id
         for parent_id in parent_id_list :
            fit_children[parent_id].add( node_id )
            node_id = parent_id
         fit_children[root_node_id].add( node_id )
   #
   # BEGIN_RETURN
   # ...
   assert type(fit_children) == list
   ok = False
   for node_id in fit_goal_set :
      ok = ok or len( fit_children[node_id] ) == 0
   assert ok
   return fit_children
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_freeze_dict}

Get Set of Covariate Multipliers That are Frozen
################################################

{xrst_literal ,
   # BEGIN_DEF , # END_DEF
}


node_table
**********
is a ``list`` of ``dict`` containing the node table for this cascade.

fit_node_id
***********
is the :ref:`create_job_table@job_table@fit_node_id` for this fit.

fit_split_reference_id
**********************
is the :ref:`create_job_table@job_table@split_reference_id` for this fit.
If this is an ``int`` ( ``None`` ),
the :ref:`option_all_table@split_covariate_name` does is (is not)
in the option_all table.

mulcov_freeze_table
*******************
is the :ref:`mulcov_freeze_table-name` for this cascade.

mulcov_freeze_dict
******************

keys
====
The keys in this dictionary is the
set of :ref:`mulcov_freeze_table@mulcov_id` for covariates
that should be frozen for jobs that are children of this fit.
In other words, if this fit job, and any parent of this fit job,
is in the mulcov_freeze table,
the corresponding *mulcov_id* is in *mulcov_freeze_dict* .

values
======
#. If this fit job is in the mulcov_freeze table with this *mulcov_id* ,
   *mulcov_freeze_dict* [ *mulcov_id* ] is 'posterior' .

#. If a parent of this fit job is in the mulcov_freeze table  with this
   *mulcov_id* , *mulcov_freeze_dict* [ *mulcov_id* ] is 'prior' .

{xrst_end get_freeze_dict}
'''
# BEGIN_DEF
# at_cascade.get_freeze_dict
def get_freeze_dict(
   node_table,
   fit_node_id,
   fit_split_reference_id,
   mulcov_freeze_table,
) :
   assert type(node_table) == list
   assert type(fit_node_id) == int
   assert type(fit_split_reference_id) == int or fit_split_reference_id == None
   assert type(mulcov_freeze_table) == list
   # END_DEF
   #
   # mulcov_freeze_dict
   mulcov_freeze_dict = dict()
   #
   # freeze_row
   for freeze_row in mulcov_freeze_table :
      #
      # freeze_split_reference_id
      freeze_split_reference_id = freeze_row['split_reference_id']
      if freeze_split_reference_id == fit_split_reference_id :
         #
         # freeze_node_id, mulcov_id
         freeze_node_id    = freeze_row['fit_node_id']
         freeze_mulcov_id  = freeze_row['mulcov_id']
         #
         # node_id
         node_id = fit_node_id
         while node_id != None :
            if freeze_node_id == node_id :
               if node_id == fit_node_id :
                  mulcov_freeze_dict[freeze_mulcov_id] = 'posterior'
               else :
                  mulcov_freeze_dict[freeze_mulcov_id] = 'prior'
            #
            # node_id
            node_id = node_table[node_id]['parent']
   #
   # BEGIN_RETURN
   assert type(mulcov_freeze_dict) == dict
   return mulcov_freeze_dict
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import csv
"""
{xrst_begin csv.join_file}

Join The Tables In Two Csv Files
################################

Prototype
*********
{xrst_literal
   BEGIN_DEF
   END_DEF
}

left_file
*********
Is the name of the first csv files.
Its columns will come first in the output file
(and in the same order as in *left_file* ).

right_file
**********
Is the name of the second csv file.
Its columns will come second in the output file
(and in the same order as in *right_file* ).
This file must have the exact same number or rows as *left_file* .

result_file
***********
Is the name of the resulting file.
It will have all the columns in *left_file* and *right_file*.
The values in the *i* -th row of *result_file* will be the
corresponding values in the *i* -th row of *left_file* and *right_file*.
If a column appears in both files,
the position of the column is its position in *left_file*,
and its value is the value in  *right_file* .

Example
*******
:ref:`csv.join_file_xam-name`


{xrst_end csv.join_file}
"""
import at_cascade
#
# BEGIN_DEF
# at_cascade.csv.join_file
def join_file(
   left_file    ,
   right_file   ,
   result_file  ,
) :
   assert type(left_file)  == str
   assert type(right_file)  == str
   assert type(result_file)  == str
   # END_DEF
   #
   # left_table, right_table
   left_table   = at_cascade.csv.read_table(left_file)
   right_table  = at_cascade.csv.read_table(right_file)
   if len(left_table) != len(right_table) :
      left_lines  = len(left_table) + 1
      right_lines = len(right_table) + 1
      msg  = 'join_file: left and right files have different number of lines\n'
      msg += f'left file  {left_file} has {left_lines}'
      msg += f'right file {right_file} has {right_lines}'
      assert False, msg
   #
   # result_table
   result_table = list()
   for index in range( len(left_table) ) :
      row  = left_table[index]
      row.update( right_table[index] )
      result_table.append( row )
   #
   # result_file
   at_cascade.csv.write_table(
      file_name = result_file  ,
      table     = result_table ,
   )



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin check_cascade_node}
{xrst_spell
  avgint
}

Check the Cascade Results for a Node
####################################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

rate_true
*********
This argument
is a function with the following syntax
```
rate = rate_true(rate_name, a, t, n, c)
```

rate_name
=========
The argument *rate_name* is one of the following:
:ref:`glossary@iota`,
:ref:`glossary@rho`,
:ref:`glossary@chi`, or
:ref:`glossary@omega`.

a, t
====
The argument *a* ( *t* ) is the age ( time ) at which we
are evaluating the rate.

n
=
The argument *n* is the :ref:`glossary@node_name` for the
node where we are evaluating the rate.

c
=
The argument *c* is a list of covariate values
in the same order as the covariate table.

rate
====
The result *rate* is the corresponding value for the rate.

all_node_database
*****************
is a python string specifying the location of the
:ref:`all_node_db-name`
relative to the current working directory.

fit_database
************
is a python string specifying the location of a
dismod_at database relative to the current working directory.
It is a :ref:`glossary@fit_database` with the
extra properties listed under
:ref:`cascade_root_node@Output dismod.db`
in the cascade_root_node documentation.

avgint_table
************
This an avgint table specifying the predictions to check.
The node_id in this table does not matter because the parent node
in the fit_database is used in its place.
The avgint table in the *fit_database* is replaced using this argument.

relative_tolerance
******************
Is an upper bound for the relative error in the predictions
corresponding to the avgint table.
The error is the prediction using the true rates
minus the prediction using the estimated rates.
If this argument is ``None``, the maximum relative error is printed.
This is intended as an aid in setting the relative tolerance.

{xrst_end check_cascade_node}
'''
# ----------------------------------------------------------------------------
import copy
import math
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.check_cascade_node
def check_cascade_node(
         rate_true          ,
         all_node_database  ,
         fit_database  ,
         avgint_table       ,
         relative_tolerance = None,
) :
   assert type(all_node_database) == str
   assert type(fit_database) == str
   assert type(avgint_table) == list
   assert relative_tolerance == None or type(relative_tolerance) == float
   # END_DEF
   #
   # root_database
   connection = dismod_at.create_connection(
      all_node_database, new = False, readonly = True
   )
   option_all_table = dismod_at.get_table_dict(connection, 'option_all')
   connection.close()
   root_database      = None
   for row in option_all_table :
      if row['option_name'] == 'root_database' :
         root_database      = row['option_value']
   assert root_database != None
   #
   # tables
   fit_or_root = at_cascade.fit_or_root_class(
      fit_database, root_database
   )
   tables = dict()
   for name in [
      'age',
      'covariate',
      'time',
      'integrand',
      'node',
   ] :
      tables[name] = fit_or_root.get_table(name)
   fit_or_root.close()
   #
   # fit_node_id
   fit_node_name = at_cascade.get_parent_node( fit_database )
   fit_node_id   = at_cascade.table_name2id(
      tables['node'], 'node', fit_node_name
   )
   #
   # avgint table
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   message         = 'check_cascade_node: replace avgint table'
   tbl_name        = 'avgint'
   avgint_copy     = copy.copy( avgint_table )
   for row in avgint_copy :
      row['node_id'] = fit_node_id
   dismod_at.replace_table(connection, tbl_name, avgint_copy)
   at_cascade.add_log_entry(connection, message)
   connection.close()
   #
   # predict_fit_var_table
   command = [ 'dismod_at', fit_database, 'predict', 'fit_var' ]
   dismod_at.system_command_prc(command)
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = True
   )
   predict_fit_var_table = dismod_at.get_table_dict(connection, 'predict')
   connection.close()
   #
   # predict_sample_table
   command = [ 'dismod_at', fit_database, 'predict', 'sample' ]
   dismod_at.system_command_prc(command)
   connection = dismod_at.create_connection(
      fit_database, new = False, readonly = True
   )
   predict_sample_table = dismod_at.get_table_dict(connection, 'predict')
   connection.close()
   #
   # n_covariate n_avgint, n_predict, n_sample
   n_covariate = len(tables['covariate'])
   n_avgint    = len(avgint_table)
   n_predict   = len(predict_sample_table)
   n_sample    = int( n_predict / n_avgint )
   #
   assert n_avgint == len( predict_fit_var_table )
   assert n_predict % n_avgint == 0
   #
   # cov_reference_list
   cov_reference_list = list()
   for (covariate_id, row) in enumerate( tables['covariate'] ) :
      reference = row['reference']
      assert reference is not None
      cov_reference_list.append( reference )
   #
   # rate_fun
   rate_fun = dict()
   rate_fun['iota'] = lambda a, t :  \
         rate_true('iota', a, t, fit_node_name, cov_reference_list)
   rate_fun['rho'] = lambda a, t :  \
         rate_true('rho', a, t, fit_node_name, cov_reference_list)
   rate_fun['chi'] = lambda a, t :  \
         rate_true('chi', a, t, fit_node_name, cov_reference_list)
   rate_fun['omega'] = lambda a, t :  \
         rate_true('omega', a, t, fit_node_name, cov_reference_list)
   #
   # sumsq
   sumsq = n_avgint * [0.0]
   #
   # predict_id, predict_row
   max_rel_error    = 0.0
   for (predict_id, predict_row) in enumerate(predict_sample_table ) :
      #
      # avgint_row
      avgint_id  = predict_row['avgint_id']
      avgint_row = avgint_copy[avgint_id]
      assert avgint_id == predict_id % n_avgint
      #
      # sample_index
      sample_index = predict_row['sample_index']
      assert sample_index * n_avgint + avgint_id == predict_id
      #
      # integrand_name
      integrand_id = avgint_row['integrand_id']
      integrand_name = tables['integrand'][integrand_id]['integrand_name']
      #
      # node_name
      node_id   = avgint_row['node_id']
      node_name = tables['node'][node_id]['node_name']
      assert node_name == fit_node_name
      #
      # avg_integrand
      avg_integrand = predict_fit_var_table[avgint_id]['avg_integrand']
      #
      # sample_value
      sample_value = predict_row['avg_integrand']
      #
      # sumsq
      sumsq[avgint_id] += (sample_value - avg_integrand)**2
   #
   # avgint_id, row
   for (avgint_id, row) in enumerate(predict_fit_var_table) :
      assert avgint_id == row['avgint_id']
      #
      # avgint_row
      avgint_row = avgint_copy[avgint_id]
      #
      # integrand_name
      integrand_id   = avgint_row['integrand_id']
      integrand_name = tables['integrand'][integrand_id]['integrand_name']
      #
      # age
      age = avgint_row['age_lower']
      assert age == avgint_row['age_upper']
      #
      # time
      time = avgint_row['time_lower']
      assert time == avgint_row['time_upper']
      #
      # avg_integrand
      avg_integrand = row['avg_integrand']
      #
      # sample_std
      sample_std = math.sqrt( sumsq[avgint_id] )
      #
      # check_value
      grid        = { 'age' : [age], 'time' : [time] }
      abs_tol     = 1e-8
      check_value = dismod_at.average_integrand(
         rate_fun, integrand_name, grid, abs_tol
      )
      #
      # check
      if check_value == 0.0 :
         assert avg_integrand == 0.0
      else :
         error            = check_value - avg_integrand
         rel_error        = 1.0 - avg_integrand / check_value
         max_rel_error    = max(max_rel_error, abs(rel_error) )
         if not relative_tolerance is None :
            if abs(rel_error) >= relative_tolerance :
               print(rel_error, relative_tolerance)
            assert abs(rel_error) < relative_tolerance
            assert abs(error) < 2.0 * sample_std
   if relative_tolerance is None :
      print('check_cascade: max_rel_error = ', max_rel_error)



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin empty_directory}

Create an Empty Directory
#########################

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

directory
*********
If this directory exists, it is removed and then re-created.
Otherwise, it is just created.


{xrst_end empty_directory}
'''
# -----------------------------------------------------------------------------
import os
import shutil
#
# BEGIN_DEF
# at_cascade.empty_directory
def empty_directory(directory) :
   assert type(directory) == str
   assert directory.startswith('build/')
   # END_DEF
   #
   if os.path.exists( directory ) :
      shutil.rmtree( directory )
   os.makedirs(directory)



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin copy_other_tbl}

Copy Root Node Database
#######################

2DO
***
This routine should be moved
from at_cascade.copy_other_tbl to dismod_at.copy_other_tbl .

Prototype
*********
{xrst_literal
   # BEGIN_DEF
   # END_DEF
}

Purpose
*******
This routine copies all the dismod_at other tables from the other database
to the dismod_at *fit_database* .
In addition, it changes the option table in the *fit_database*
so that it does not require any tables in another database.

{xrst_end copy_other_tbl}
'''
import os
import shutil
import tempfile
import at_cascade
import dismod_at
#
# BEGIN_DEF
# at_cascade.copy_other_tbl
def copy_other_tbl(fit_database) :
   assert type(fit_database) == str
   # END_DEF
   #
   # fit_connection
   fit_connection = dismod_at.create_connection(
      fit_database, new = False, readonly = False
   )
   #
   # option_table
   option_table = dismod_at.get_table_dict(fit_connection, tbl_name = 'option')
   #
   # other_database, other_input_table, option_table
   other_database    = None
   other_input_table = None
   for row in option_table :
      if row['option_name'] == 'other_database' :
         other_database      = row['option_value']
         row['option_value'] = None
      if row['option_name'] == 'other_input_table' :
         other_input_table   = row['option_value']
         row['option_value'] = None
   #
   # other_table_list
   other_table_list = other_input_table.split()
   #
   # fit_database.option_table
   dismod_at.replace_table(fit_connection, 'option', option_table)
   #
   # other_database
   index = fit_database.rindex('/')
   if 0 <= index :
      fit_dir        = fit_database[0 : index]
      other_database = f'{fit_dir}/{other_database}'
   #
   # other_connection
   other_connection = dismod_at.create_connection(
      other_database, new = False, readonly = True
   )
   #
   # table_name
   for tbl_name in other_table_list :
      #
      table              = dismod_at.get_table_dict(other_connection, tbl_name)
      col_name, col_type = dismod_at.get_name_type(other_connection,  tbl_name)
      row_list           = list()
      primary_key        = f'{tbl_name}_id'
      assert col_name[0] == primary_key
      col_name.pop(0)
      col_type.pop(0)
      for row_in in table :
         row_out = list()
         for name in col_name :
            row_out.append( row_in[name] )
         row_list.append( row_out )
      dismod_at.create_table(
         fit_connection, tbl_name, col_name, col_type, row_list
      )
   #
   # fit_connection, other_connection
   fit_connection.close()
   other_connection.close()



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2024 Bradley M. Bell
# ----------------------------------------------------------------------------
import copy
'''
{xrst_begin csv.covariate_both}

Add both Sex to csv Covariate Table
###################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

covariate_table_in
******************
Is a ``list`` of ``dict`` representation of a
:ref:`csv.simulate@Input Files@covariate.csv` file.

covariate_table_out
*******************
Is a ``list`` of ``dict`` representation of
:ref:`csv.simulate@Input Files@covariate.csv` file.
This has all the rows that are in *covariate_table_in*
plus rows with *sex* equal to *both* .

both
====
For each node_name, age, time,
the covariate table value for *sex* equal ``both``
is the average of its value for ``female`` and ``male`` .
To be specific, if female_value, male_value, both_value
are a covariate's values for a specific (node_name, age, time) then:
{xrst_code py}
   if female_value == male_value :
      both_value = female_value
   else :
      both_value = ( float( female_value ) + float( male_value ) ) / 2.0
{xrst_code}

Example
*******
see :ref:`csv.cov_both_xam-name` .

{xrst_end csv.covariate_both}
'''
# BEGIN_DEF
# at_cascade.csv.covariate_both
def covariate_both(covariate_table_in) :
   assert type(covariate_table_in) == list
   for row in covariate_table_in :
      for key in row :
         type_check = str if key in [ 'node_name' , 'sex' ] else float
         assert type( row[key] ) == type_check
   # END_DEF
   #
   # covariate_dict
   covariate_dict = dict()
   covariate_dict['female'] = dict()
   covariate_dict['male']   = dict()
   for row in covariate_table_in :
      sex = row['sex']
      if sex not in [ 'female', 'male' ] :
         msg = 'covariate_both: covariate.csv: sex is not female or male'
         assert False, msg
      triple  = ( row['node_name'], row['age'], row['time'] )
      covariate_dict[sex][triple] = row
   for triple in covariate_dict['female'] :
      if triple not in covariate_dict['male'] :
         (node_name, age, time) = triple
         msg  = f'covariate_both: covariate.csv: found '
         msg += f'node_name = {node_name}, age = {age}, time = {time}\n'
         msg += 'for sex = female but not for sex = male'
         assert False, msg
   for triple in covariate_dict['male'] :
      if triple not in covariate_dict['female'] :
         (node_name, age, time) = triple
         msg  = f'covariate_both: covariate.csv: found '
         msg += f'node_name = {node_name}, age = {age}, time = {time}\n'
         msg += 'for sex = male but not for sex = female'
         assert False, msg
   #
   # cov_list
   exclude    = { 'node_name', 'sex', 'age', 'time'}
   cov_list   = list( set( covariate_table_in[0].keys() ) - exclude )
   #
   # covariate_table_out
   covariate_table_out = copy.deepcopy(covariate_table_in)
   for triple in covariate_dict['female'] :
      (node_name, age, time) = triple
      row_female  = covariate_dict['female'][triple]
      row_male       = covariate_dict['male'][triple]
      #
      assert row_female['sex'] == 'female'
      assert row_male['sex']   == 'male'
      #
      assert row_female['node_name'] == node_name
      assert row_male['node_name']   == node_name
      #
      assert row_female['age'] == age
      assert row_male['age']   == age
      #
      assert row_female['time'] == time
      assert row_male['time']   == time
      #
      row_out        = copy.copy( row_female )
      row_out['sex'] = 'both'
      for key in cov_list :
         if row_male[key] != row_female[key] :
            row_out[key] = (row_female[key] + row_male[key]) / 2.0
      covariate_table_out.append( row_out )
   # BEGIN_RETURN
   #
   assert type(covariate_table_out) == list
   for row in covariate_table_out :
      for key in row :
         type_check = str if key in [ 'node_name' , 'sex' ] else float
         assert type( row[key] ) == type_check
   return covariate_table_out
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-26 Bradley M. Bell
# ----------------------------------------------------------------------------
import numpy
import scipy.interpolate
"""
{xrst_begin bilinear}

Bilinear Spline Interpolation
#############################

Prototype
*********
{xrst_literal ,
   BEGIN_DEF, END_DEF
   BEGIN_RETURN, END_RETURN
}

x_name
******
is the name in *table* for the x
variable in the interpolation.

y_name
******
is the name in *table* for the y
variable in the interpolation.

z_list
******
Each element of  *z_list* is a ``str`` specifying
the name in *table* for a z
variable in the interpolation.

table
*****
Each element of *table* is a ``dict``.
The values *x_name* , *y_name* and each element of *z_list*
must be a key in each of the dictionaries.
Furthermore, the corresponding dictionary values must be ``str`` or ``float`` .
This table must be non-empty.

Rectangular Grid
================
For each x in *x_grid* and y in *y_grid* there must be one,
and only one, row in *table* with that x and y value.
If this is not the case, ``bilinear`` will return with *spline_dict*
equal to ``None`` .

x_grid
******
Is the a sorted ``list`` of values that appear
in the *x_name* column of *table* .
Duplicates are not included in this list.

y_grid
******
Is the a sorted ``list`` of values that appear
in the *y_name* column of *table* .
Duplicates are not included in this list.

spline_dict
***********
Is a ``dict`` of spline functions with keys equal to
each *z_name* in  *z_list* .
The function call
{xrst_code py}
   z = spline_dict[z_name](x, y)
{xrst_code}
sets z to the value of the spline for *z_name* .

#. The values x and y are ``float`` or ``int`` .
#. The value z is a  ``float`` .
#. The function is extended as constant with respect to x (y)
   for values of x (y) outside the limits of x_grid (y_grid).

Example
*******
:ref:`example_bilinear-name`

{xrst_end bilinear}
"""
class spline_wrapper :
   def __init__(self, spline, box, const_x, const_y) :
      self.spline  = spline
      self.box     = box
      self.const_x = const_x
      self.const_y = const_y
   def __call__(self, x, y) :
      #
      # x, y
      if type(x) == int :
         x = float(x)
      if type(y) == int :
         y = float(y)
      assert type(x) == float
      assert type(y) == float
      #
      # x, y
      # The documentation for RectBivariateSpline says
      # 'Evaluated points outside the data range will be extrapolated.' .
      # but testing indicates the following is not necessary
      x = max(x, self.box['x_min'])
      x = min(x, self.box['x_max'])
      y = max(y, self.box['y_min'])
      y = min(y, self.box['y_max'])
      #
      # result
      if self.const_x and self.const_y :
         result = self.spline
      elif self.const_x :
         result = self.spline(y)
      elif self.const_y :
         result = self.spline(x)
      else :
         result = self.spline(x, y)
      assert type(result) == numpy.ndarray
      assert result.size == 1
      #
      result = result.flatten()
      result = result[0]
      return result

# BEGIN_DEF
# at_cascade.bilinear
def bilinear(
   table,
   x_name,
   y_name,
   z_list
) :
   assert type(table) == list
   assert type(x_name) == str
   assert type(y_name) == str
   assert type(z_list) == list
   # END_DEF
   #
   if len(table) == 0 :
      return (list(), list(), None)
   #
   # x_set, y_set
   x_set       = set()
   y_set       = set()
   for row in table :
      x      = float( row[x_name] )
      y      = float( row[y_name] )
      x_set.add(x)
      y_set.add(y)
   #
   # triple_list
   triple_list = list()
   for row in table :
      x      = float( row[x_name] )
      y      = float( row[y_name] )
      triple = (x, y, row)
      triple_list.append( triple )
      #
   #
   # n_x, n_y
   n_x = len(x_set)
   n_y = len(y_set)
   #
   # x_grid, y_grid
   x_grid = sorted(x_set)
   y_grid = sorted(y_set)
   #
   if len(triple_list) != n_x * n_y :
      return (x_grid, y_grid, None)
   #
   # triple_list
   triple_list = sorted(triple_list)
   #
   # spline_dict
   spline_dict = dict()
   #
   # z_name
   for z_name in z_list :
      #
      # z_grid
      z_grid    = numpy.empty( (n_x, n_y) )
      z_grid[:] = numpy.nan
      #
      # index, triple
      for (index, triple) in enumerate( triple_list ) :
         #
         # x_index, y_index
         x        = triple[0]
         y        = triple[1]
         x_index  = int( index / n_y )
         y_index  = index % n_y
         if x != x_grid[x_index] :
            return(x_grid, y_grid, None)
         if y != y_grid[y_index] :
            return(x_grid, y_grid, None)
         #
         # z_grid
         row                      = triple[2]
         value                    = float( row[z_name] )
         z_grid[x_index, y_index] =  value
      #
      # const_x
      const_x = True
      for y_index in range(n_y) :
         const_col = numpy.all( z_grid[0, y_index] == z_grid[:, y_index] )
         const_x   = const_x and const_col
      #
      # const_y
      const_y = True
      for x_index in range(n_x) :
         const_row = numpy.all( z_grid[x_index, 0] == z_grid[x_index, :] )
         const_y   = const_y and const_row
      #
      # spline
      if const_x and const_y :
         spline = numpy.array( z_grid[0, 0] )
      elif const_x :
         spline = scipy.interpolate.UnivariateSpline(
            y_grid, z_grid[0,:], k=1, s=0
         )
      elif const_y :
         spline = scipy.interpolate.UnivariateSpline(
            x_grid, z_grid[:,0], k=1, s=0
         )
      else :
         spline = scipy.interpolate.RectBivariateSpline(
            x_grid, y_grid, z_grid, kx=1, ky=1, s=0
         )
      box = {
         'x_min' : x_grid[0],
         'x_max' : x_grid[-1],
         'y_min' : y_grid[0],
         'y_max' : y_grid[-1],
      }
      spline_dict[z_name] = spline_wrapper(spline , box, const_x, const_y)
   #
   # BEGIN_RETURN
   # ...
   assert type(x_grid) == list
   assert type(y_grid) == list
   assert type(spline_dict) == dict
   return (x_grid, y_grid, spline_dict)
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin csv.module}

The at_cascade.csv Python Module
################################

Interface Example
*****************
The python code that implements this interface to the at_cascade
:ref:`module-name` can be used as an example for other interfaces.

Notation
********

Demographer
===========
None of the data is in demographer notation.
For example,
:ref:`csv.simulate@Input Files@covariate.csv@time`
1990 means the beginning of 1990,
not the time interval from 1990 to 1991.

Rectangular Grid
================
Define a selection subset of a csv file as those rows with the same
value in column *select*.
For each value in *select* ,
a csv file is said to have a rectangular grid in age and time
if the following holds:

#. Define :math:`( a_1 , \ldots , a_M )`
   to be the vector of values in the age column for this selection subset.

#. Define :math:`( t_1 , \ldots , t_N )`
   to be the vector of values in the time column for this selection subset.

#. For :math:`m = 1 , \ldots , M`, :math:`n = 1,  \ldots , N` ,
   there is one and only one row in this selection subset with
   age equal to :math:`a_m`, and
   time equal to :math:`t_n`.

Data Type
=========
The actual data type for each entry in a csv file is a string; i.e.,
an arbitrary sequence of characters. Certain columns have further
restrictions as described below

Integer
-------
An integer value is a string represents of an integer.

Float
-----
A float value is a string that represents a floating point number.

Sex
---
A sex, or sex_name, is one of the following:
``female`` , ``both``, or ``male``.

Boolean
-------
A boolean value is either ``true`` or ``false`` .

Index Column
============
An index column for a csv file is an integer column
that has the row number corresponding to each row.
It starts with zero at the first row below the header row.
If a column name is an index column for two or more files,
rows with the same index value in the different files
correspond to each other.

Distributions
=============
Unless other wise specified, the mean and standard deviations that
simulate refers to are for a normal distribution.

sex_name2value
**************
The following dictionary maps each sex name to the corresponding sex value
{xrst_code py}'''
sex_name2value = { 'female' : -0.5, 'both' : 0.0, 'male' : 0.5 }
'''{xrst_code}

split_reference_table
*********************
The :ref:`option_all_table@split_covariate_name` is ``sex`` and
the :ref:`split_reference_table-name` has the following values for
csv.fit, :ref:`csv.simulate-name` and :ref:`csv.predict-name` :
{xrst_code py}'''
split_reference_table = [
   { 'split_reference_name' : 'female' , 'split_reference_value' : -0.5 },
   { 'split_reference_name' : 'both'   , 'split_reference_value' :  0.0 },
   { 'split_reference_name' : 'male'   , 'split_reference_value' : +0.5 },
]
'''{xrst_code}

Routines
********

.. BEGIN_SORT_THIS_LINE_PLUS_2
{xrst_toc_table
   at_cascade/csv/ancestor_fit.py
   at_cascade/csv/check_table.py
   at_cascade/csv/covariate_avg.py
   at_cascade/csv/covariate_both.py
   at_cascade/csv/covariate_same.py
   at_cascade/csv/covariate_spline.py
   at_cascade/csv/empty_str.py
   at_cascade/csv/fit.py
   at_cascade/csv/get_header.py
   at_cascade/csv/join_file.py
   at_cascade/csv/pre_one_job.py
   at_cascade/csv/pre_one_process.py
   at_cascade/csv/pre_parallel.py
   at_cascade/csv/pre_user.py
   at_cascade/csv/predict.py
   at_cascade/csv/read_table.py
   at_cascade/csv/set_truth.py
   at_cascade/csv/simulate.py
   at_cascade/csv/write_table.py
}
.. END_SORT_THIS_LINE_MINUS_2

{xrst_end csv.module}
'''
# BEGIN_SORT_THIS_LINE_PLUS_1
from .ancestor_fit     import ancestor_fit
from .check_table      import check_table
from .covariate_avg    import covariate_avg
from .covariate_both   import covariate_both
from .covariate_same   import covariate_same
from .covariate_spline import covariate_spline
from .empty_str        import empty_str
from .fit              import fit
from .get_header       import get_header
from .join_file        import join_file
from .pre_one_job      import pre_one_job
from .pre_one_process  import pre_one_process
from .pre_parallel     import pre_parallel
from .pre_user         import pre_user
from .predict          import predict
from .read_table       import read_table
from .set_truth        import set_truth
from .simulate         import simulate
from .write_table      import write_table
# END_SORT_THIS_LINE_MINUS_1



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import csv
"""
{xrst_begin csv.get_header}

Get the Header from a CSV File
##############################

Prototype
*********
{xrst_literal ,
   BEGIN_DEF, END_DEF
   BEGIN_RETURN, END_RETURN
}


file_name
*********
is a ``str`` with the name of the CSV file.
The first line of the file is the header line and the others contain
the data.
The header value in the j-th column of the first line is the
corresponding column name.

header
******
the return value *header* is a ``list`` of ``str``.
We use *n* to denote the length of the list which is
the number of commas in the first line plus one.
For *j*, an ``int`` ,  between zero and *n* -1,
the name of the *j*\ -th column is *header* [ *j* ].

Example
*******
:ref:`csv.table-name`

{xrst_end csv.get_header}
"""
# BEGIN_DEF
# at_cascade.csv.get_header
def get_header(file_name) :
   assert type(file_name)  == str
   # END_DEF
   #
   file_ptr   = open(file_name)
   reader     = csv.reader(file_ptr)
   header     = next( reader )
   #
   # BEGIN_RETURN
   # ...
   assert type(header) == list
   return header
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_parent_node}

Get Parent Node Name From Option Table
######################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN,  # END_RETURN
}

database
********
This is a dismod_at database.

parent_node_name
****************
is the *database* option table
*option_value* for the row where *option_name* is ``parent_node_name``.
An assert will be generated if no such *option_name* is found in the
option table.


{xrst_end get_parent_node}
'''
import dismod_at
# ----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.get_parent_node
def get_parent_node(database = None) :
   assert type(database) == str
   # END_DEF
   #
   # option_table
   connection       = dismod_at.create_connection(
      database, new = False, readonly = True
   )
   option_table     = dismod_at.get_table_dict(connection, 'option')
   connection.close()
   #
   # parent_node_name
   parent_node_name = None
   for row in option_table :
      if row['option_name'] == 'parent_node_name' :
         parent_node_name = row['option_value']
   msg = 'Cannot find parent_node_name row in option table in ' + database
   assert not parent_node_name is None, msg
   #
   # BEGIN_RETURN
   # ..
   assert type(parent_node_name) == str
   return parent_node_name
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin csv.ancestor_fit}

Determine Closet Ancestor With Fit and Samples
##############################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

fit_dir
*******
is the directory where the csv files are located.

job_table
*********
is the :ref:`create_job_table@job_table` for this cascade.

predict_job_id
**************
is the :ref:`create_job_table@job_table@job_id` for this prediction.

node_table
**********
is the list of dict corresponding to the node table for this cascade.

root_node_id
************
is the node_id in the node table for the root node for this cascade.
Note that csv version of the cascade does its sex split at this node.

root_split_reference_id
***********************
is the split_reference_id (sex id) for the root node of the cascade.
The cascade can begin at female, both, or male.

at_cascade_log_dict
*******************
is a dictionary, with keys equal to job names, containing
the log messages that have type ``at_cascade`` .
The messages for each key are in the log table for the corresponding job.

allow_same_job
**************
If this is true (false) the a job corresponding to the ancestor fit
can be the same as the job for the prediction; i.e., the predict job
is the closest ancestor job. Otherwise, the parent of the predict job
is the closest ancestor job.

predict_job_dir
***************
This is the directory, relative to the *fit_dir*,
that corresponds to the *predict_job_id* .
See :ref:`get_database_dir-name` .

ancestor_job_dir
****************
This is the directory, relative to the *fit_dir*,
that corresponds to the closest ancestor of *predict_job_id*
that had a successful fit and posterior sampling.
To be specific, ``sample: OK`` is in its
:ref:`fit_one_job@fit_database@log` .


{xrst_end csv.ancestor_fit}
'''
import os
import at_cascade

# BEGIN_DEF
# at_cascade.csv.ancestor_fit
def ancestor_fit(
   fit_dir,
   job_table,
   predict_job_id,
   node_table,
   root_node_id,
   split_reference_table,
   root_split_reference_id,
   at_cascade_log_dict,
   allow_same_job,
) :
   assert type(fit_dir) == str
   assert type(job_table) == list
   assert type(predict_job_id) == int
   assert type(node_table) == list
   assert type( root_node_id ) == int
   assert type(split_reference_table) == list
   assert type( root_split_reference_id) == int
   assert type( at_cascade_log_dict ) == dict
   assert type( allow_same_job ) == bool
   # END_DEF
   #
   # node_split_set
   node_split_set = { root_node_id }
   #
   # job_name, predict_node_id, predict_split_reference_id
   job_row                     = job_table[predict_job_id]
   job_name                    = job_row['job_name']
   predict_node_id             = job_row['fit_node_id']
   predict_split_reference_id  = job_row['split_reference_id']
   #
   # predict_job_dir, ancestor_job_dir
   predict_job_dir = at_cascade.get_database_dir(
      node_table              = node_table                     ,
      split_reference_table   = split_reference_table          ,
      node_split_set          = node_split_set                 ,
      root_node_id            = root_node_id                   ,
      root_split_reference_id = root_split_reference_id        ,
      fit_node_id             = predict_node_id                ,
      fit_split_reference_id  = predict_split_reference_id     ,
   )
   #
   # sample_ok
   predict_node_database = f'{fit_dir}/{predict_job_dir}/dismod.db'
   sample_ok = False
   if os.path.exists( predict_node_database ) :
      messages  = at_cascade_log_dict[job_name]
      sample_ok  = 'sample: OK' in messages
   if sample_ok and allow_same_job :
      ancestor_job_dir = predict_job_dir
      return predict_job_dir, ancestor_job_dir
   sample_ok = False
   #
   # job_id, ancestor_job_dir
   job_id            = predict_job_id
   while not sample_ok :
      #
      # job_id
      job_id = job_table[job_id]['parent_job_id']
      if job_id == None :
         ancestor_job_dir = None
         assert type(predict_job_dir) == str
         return predict_job_dir, ancestor_job_dir
      #
      # job_name, ancestor_node_id, ancestor_split_reference_id
      job_row                      = job_table[job_id]
      job_name                     = job_row['job_name']
      ancestor_node_id             = job_row['fit_node_id']
      ancestor_split_reference_id  = job_row['split_reference_id']
      #
      # ancestor_job_dir
      ancestor_job_dir = at_cascade.get_database_dir(
         node_table              = node_table                     ,
         split_reference_table   = split_reference_table          ,
         node_split_set          = node_split_set                 ,
         root_node_id            = root_node_id                   ,
         root_split_reference_id = root_split_reference_id        ,
         fit_node_id             = ancestor_node_id               ,
         fit_split_reference_id  = ancestor_split_reference_id     ,
      )
      #
      # sample_ok
      ancestor_job_database = f'{fit_dir}/{ancestor_job_dir}/dismod.db'
      if os.path.exists( ancestor_job_database ) :
         messages   = at_cascade_log_dict[job_name]
         sample_ok  = 'sample: OK' in messages
   #
   # BEGIN_RETURN
   assert type(predict_job_dir) == str
   if type(ancestor_job_dir) == str :
      assert predict_job_dir.startswith(ancestor_job_dir)
   else :
      assert ancestor_job_dir == None
   #
   return predict_job_dir, ancestor_job_dir
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r'''
{xrst_begin table_name2id}
{xrst_spell
  tbl
}

Map a Table Row Name to The Row Index
#####################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, END_DEF
   # BEGIN_RETURN, END_RETURN
}

table
*****
This is a ``list`` of ``dict`` containing the table.
The primary key is not included because it is the row index.

tbl_name
********
This is a ``str`` containing the name of the table.
This table must have a column named *table_name*\ ``_name``.

row_name
********
This is the value we are searching for in the
*table_name*\ ``_name`` column.

row_id
******
This is the index of the row in the table where
*row_name* occurs. An assert will occur if there is no such row.

{xrst_end table_name2id}
'''
# -----------------------------------------------------------------------------
# BEGIN_DEF
# at_cascade.table_name2id
def table_name2id(
   table, tbl_name, row_name
) :
   assert type(table) == list
   assert type(tbl_name) == str
   # END_DEF
   col_name = tbl_name + '_name'
   row_id   = None
   for (index, row) in enumerate(table) :
      if row[col_name] == row_name :
         row_id = index
   if row_id == None :
      msg  = f'table_name2id: "{row_name}" '
      msg += f'is not presnet in column "{col_name}" of "{tbl_name}" table.'
      assert False, msg
   #
   # BEGIN_RETURN
   # ...
   assert type(row_id) == int
   return row_id
   # END_RETURN



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin fit_or_root_class}

Get Tables Corresponding to a Fit Node
######################################

fit_or_root_class
*****************
{xrst_code py}
fit_or_root = fit_or_root_class(fit_database, root_database)
{xrst_code}

fit_database
============
This ``str`` is the name of a :ref:`glossary@fit_database` .

root_database
=============
This ``str`` is the name of a :ref:`glossary@root_database` .

get_table
*********
{xrst_code py}
table = fit_or_root.get_table(table_name)
{xrst_code}

table_name
==========
This ``str`` the name of the table that we are getting.

table
=====
This is a ``list`` of ``dict`` representation of the table.
If *table_name* is a
:ref:`constant table <module@at_cascade.constant_table_list>` ,
*table* is retrieved from the root node database.
Otherwise it is retrieved from the fit node database.

null_row
********
{xrst_code py}
row = fit_or_root.null_row(table_name)
{xrst_code}

table_name
==========
This ``str`` the name of the table that we getting a null row for.

row
===
is a ``dict`` with one key for every column in the table specified
by *table_name* and all the values equal to ``None`` .


close
*****
{xrst_code py}
fit_or_root.close()
{xrst_code}
This closes the database connection held by a fit_or_root.
The ``get_table`` function will no longer be available.

{xrst_end fit_or_root_class}
'''
import dismod_at
import at_cascade
#
class fit_or_root_class :
   #
   # __init__
   def __init__(self, fit_database, root_database) :
      assert type(fit_database) == str
      assert type(root_database) == str
      #
      self.fit_connection = dismod_at.create_connection(
         fit_database, new = False, readonly = True
      )
      self.root_connection = dismod_at.create_connection(
         root_database, new = False, readonly = True
      )
      self.open = True
   #
   # get_table
   def get_table(self, table_name) :
      assert type(table_name) == str
      assert self.open
      #
      if table_name in at_cascade.constant_table_list :
         table = dismod_at.get_table_dict(self.root_connection, table_name)
      else :
         table = dismod_at.get_table_dict(self.fit_connection, table_name)
      return table
   #
   # null_row
   def null_row(self, table_name) :
      if table_name in at_cascade.constant_table_list :
         connection = self.root_connection
      else :
         connection = self.fit_connection
      (col_name, col_type) = dismod_at.get_name_type(connection, table_name)
      row = dict()
      for key in col_name :
         row[key] = None
      return row
   #
   # close
   def close(self) :
      self.fit_connection.close()
      self.root_connection.close()
      self.open = False



# SPDX-License-Identifier: AGPL-3.0-or-later
# SPDX-FileCopyrightText: University of Washington <https://www.washington.edu>
# SPDX-FileContributor: 2021-24 Bradley M. Bell
# ----------------------------------------------------------------------------
'''
{xrst_begin get_fit_integrand}

Determine the Set of Integrands in Data Table
#############################################

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, END_DEF
   # BEGIN_RETURN, END_RETURN
}

fit_or_root
***********
This :ref:`fit_or_root_class-name` object has open connections
for the node we are fitting.

fit_integrand
*************
The return value *fit_integrand* is a python set of integrand_id
that appear in the data table in the *fit_database*.
Furthermore there is a row in the data table
where each such integrand_id is not held out.

{xrst_end get_fit_integrand}
'''
# ----------------------------------------------------------------------------
import sys
import dismod_at
import at_cascade
# ----------------------------------------------------------------------------
#
# BEGIN_DEF
# at_cascade.get_fit_integrand
def get_fit_integrand(fit_or_root) :
   assert type(fit_or_root) == at_cascade.fit_or_root_class
   # END_DEF
   #
   # data_table
   data_table = fit_or_root.get_table('data')
   #
   # fit_integrand
   fit_integrand = set()
   for row in data_table :
      if row['hold_out'] == 0 :
         fit_integrand.add( row['integrand_id'] )
   #
   # BEGIN_RETURN
   # ...
   assert type(fit_integrand) == set
   return fit_integrand
   # END_RETURN
